/*
   External Function Library for allocation with LP_SOLVE

   Copyright 2013 Tom Fiddaman, Ventana Systems, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0 

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/


/****************************************************************************
   Based on:
   extern.c -  Version 5.8c
   Copyright (c) 1992-2005 Ventana Systems, Inc.
   Example external function file definitions for use
   with Vensim 

  Background:
  
  At their simplest external functions are simply mathematical routines 
  that do things it is hard to do in Vensim.  In practice, however, most
  of the value of external functions comes from the ability to do 
  programming language things.  In particular loops, iterations and 
  conditional computation.  Thus there are a number of different 
  attributes an external function can have.

  1. Use only floating point numbers as inputs and output a floating point
     number.  

  2. Use a combination of floating point numbers and vectors (or arrays) 
     of values to output a floating point number.  External functions can
     also take string variables and Lookups as arguments.

  3. Take over control of for variable loops from Vensim.

  4. Directly modify arguments passed as vectors.

  Of these number 1 is straightforward and is illustrated by the COSINE 
  and INRANGE functions included here.  2 is straghtforward if you follow
  the example of the PSUM and MYMESSAGE function included here.  3 is tricky 
  because the example of the Vensim function has 1 less argument than the 
  real function, but should not cause problem if you follow example of the 
  MATRIX_INVERT function included here.

  A special note on number 2 is that you can also use this function to 
  create a simultaneous equation solver using your own solution algorithm.
  MYFINDZERO is an example of this.

  Number 4 is dangerous.  Vensim loses control of equation ordering and 
  you can get unexpected results.  Use this only if you have to.  The
  MATRIX_INPLACE_INVERT function does this.  Be careful to study the model
  (venext.mdl) shipped with these examples.

  The rest of this file is broken up into 7 sections.  If you have more than
  a few functions, you should probably take section 2 and put it into a .h 
  file.  Section 7 can be put in a separate file (or files) and the .h file 
  included in thse files.  You most likely do not need to include <windows.h>
  in the other files (but <math.h> is a good idea.  


  COMPILER NOTES

  Most of the development and testing of Vensim occurs using Microsoft
  compilers.  Visual C/C++ 

  If you want to write external functions in C plus plus you can rename this 
  file to venext.cpp and everything should work OK.

  Note that you must include venext.def to get this all to work in your 
  project (under Windows).

  CONFIGURATION NOTES

  Vensim's external functions can be used with Vensim DSS, Vensim Runtime, 
  Vensim Application Runtime and the DLLs with the exception of the minimal
  DLL. In order to have external functions loaded in Application Runtime or
  the DLLs you will need to execute the command

      SPECIAL>READINI|inifile.ini

  where inifile.ini contains the line

      ExternalFunctionLibrary=c:\program files\vensim\comp\venext.dll

  set the path appropriate for where you have located you function library.
  There is no need to call the file venext.dll - any name will do. 
  
  When you create your external function library you need to link with the
  .lib file associated with the program that will call your external functions.
  For Vensim DSS this is vensim.lib. For the Vensim DLL it is vendll32.lib
  and so on. If you do not do this loading the external functions will likely 
  fail, possibly with a message that the application is not properly installed.

    
  PLATFORM NOTES

  Linux

  Standard C files such as this should work fine under Linux using shared 
  object libraries. The sample make file venextun.mak can be used to build 
  the shared object libary - alternatively just use the command

  gcc -shared venext.c -olibvenext.so 

  because of the way Linux does runtime linking there is no need to reference
  the vensim symbols during compilation - references will be resolved at 
  runtime.

  The format for indicating where to find the external function library 
  is the same as on Windows - use the SPECIAL>READINI command. The path 
  should follow Unix naming conventions eg

     ExternalFunctionLibrary=/home/bob/newven/comp/libvenext.so

  Remember that Linux is case sensitive so it is best to keep all filenames
  lower case.
  
  One thing worth noting about Linux is that the symbol space for shared object
  libraries is common and this means that it is possible to inadvertantly 
  create functions or global variables which conflict with names in Vensim
  proper. Unfortunately, this tends to manifest itself only by causing
  catastrophic failure when the wrong function is called or the wrong
  global variable refrenced. For example, GLOB_VARS *VENGV used to be
  just GV which is used internally in Vensim and this causes a failure. 
  If you suspect this type of problem try renaming the functions or 
  variables to see if it helps.



   */

#define WANT_WINDOWS_INCLUDES /* the sample implementation of this requires windows includes/libraries */
#define VENEXT_GLOBALS
#include "vensim.h"
#ifdef unix
#include <string.h>
#include <stdlib.h>
#else
#include <malloc.h>
#endif

GLOB_VARS *VENGV='\0'; /* the value for this is set by set_gv below */



/*******************************************
 1 - function ids - used to swich between choices 
 *******************************************/
#define LP_ALLOC_FN 0
#define LP_SOLVE_FN 1

/*******************************************
 2 - function prototypes 
 *******************************************
  each external function is prototyped here
  arguments can reasonably be doubles - for normal number manipulation,
  COMPREAL * for vector manipulation or int for indexing.  Recasting of
  values takes place in 
  Note that if you use more than 1 file for the external function definitions
  you should probably put these prototypes into a #include file.  Also, for
  working with compiled simulation a # include file is helpfule, and should
  be nested into vensim.h

  Note that all the external functions are all upper case.  This is required
  if you want to use compile simulations - since the calls in mdl.c will
  be upper case.  Our apologies to those this offends.
*********************************************/

BigREAL LP_ALLOC( VECTOR_ARG *allocation, VECTOR_ARG *ds_priority, VECTOR_ARG *ds_capacity, VECTOR_ARG *d_capacity, VECTOR_ARG *s_capacity) ;
BigREAL LP( VECTOR_ARG *solution, VECTOR_ARG *objective, VECTOR_ARG *lbounds, VECTOR_ARG *ubounds, VECTOR_ARG *constraints, VECTOR_ARG *c_vals, VECTOR_ARG *c_types ) ;

/****************************************************
 3 - Grouping of functions in a structure - see venext.h
 ***********************************************************/
/*
   char *sym ; // function names - as it will be used in Vensim 
   char *argument_desc ; // description of arguments use in equation editor 
   unsigned char num_args ; // 0 - 255 
   unsigned char num_vector ; // for manipulating vectors or changing the
                                 value of model variables 
   unsigned short func_index ; // the index for the user function 0-32000 
   char num_loop  ; // the number of user managed loops - for defining
               more than one element of an array or matrix  - use -1 to signal
               a constant definition function and -2 a data definition function 
   unsigned char modify ; // flag to indicate that function may change its
                             its inputs - only makes sense if num_vector > 0 
   unsigned char num_literal ;
   unsigned char num_lookup ;
*/

static FUNC_DESC Flist[] = {
                    {"LP ALLOC"," {priority[d,s]} , {capacity[d,s]}, {capacity[d]}, {capacity[s]} ",4,4,LP_ALLOC_FN,2,0,0,0},
					{"LP SOLVE"," {objective[v]} , {lbounds[v]}, {ubounds[v]}, {constraints[c,v]}, {c_vals[c]}, {c_types[c]} ",6,6,LP_SOLVE_FN,1,0,0,0},
                    {'\0',0,0,0}} ;


/******************************************************
 4 - DLL required functions LibMain and WEP
 Obsolete - 16 bit windows only
 ******************************************************/
 
/****************************************************************
 5 External function definitions
 ****************************************************************/

 /**************************
  This function is a version check - it is required to be sure
  the external functions are compatible with the current Vensim
  version. Note that for 5.8c this number has changed but you can
  simply update this version number, the set_gv function and the
  funcversion_info function (these 2 return different value types)
  make no other changes to your external functions and they will
  work. To simplify use with different configurations (ie the 
  Model Reader and DLLs) we recommend that you not link with 
  vensim.lib or vensimdp.lib and replace

  vensim_error_message with (*VENGV->error_message)
  vensim_alloc_simmem with (*VENGV->alloc_simmem)
  vensim_execute_curloop with (*VENGV->execute_curloop)

**********************/

int VEFCC version_info()
 {
  return(EXTERNAL_VERSION) ;
  }

/* When you make changes to functions update this. If Vensim opens a .vmf file
   that references aexternal functions and there is a mismatch a message will
   be given indicating that the model should be reformed and cleaned 
   - if you return 0 no checking will occur

  If you have multiple external function libraries you can also use this to 
  signal when a model is not matched to the library (though it won't indicate
  which library should be used) */
unsigned short VEFCC funcversion_info()
{
	return(1) ;
}

int VEFCC set_gv(GLOB_VARS *vgv) 
{
	int result=0;
	VENGV = vgv ;

	if(!VENGV || 
		VENGV->vgv_magic_start != VGV_MAGIC_START ||
		VENGV->vgv_magic_end != VGV_MAGIC_END)
		return 0 ;

	result = lp_solve_load();
	if( result == 0) {
		(*VENGV->error_message)(VERROR,(unsigned char *)"Failed to load lp_solve for external function library.") ;
		return 0 ;
	}
	else if( result == -1) {
		(*VENGV->error_message)(VERROR,(unsigned char *)"Failed to load lp_solve functions.") ;
		return 0 ;
	}
	else if( result == -2) {
		(*VENGV->error_message)(VERROR,(unsigned char *)"Failed lp_solve test.") ;
		return 0 ;
	}

	return 1 ;
}

  /* ****************************
   This function is _exported and called multiple times at Vensim 
   startup with an index  - it returns 1 on success and 0 to 
   indicate the end of the function list.  For convenience the structure
   defined in section 3 is used, but all the functions could also be declared
   with a switch statement on i
 *****************************************************************/

int VEFCC user_definition(
   int i, /* an index for requesting information - this is mapped to Flist
             but could be used another way - vensim repeatedly calls
             user_definition with i bigger by 1 until user_definition returns
             0 */
   unsigned char **sym,/* the name of the function to be used in the Vensim model */
   unsigned char **argument_desc, /* description of arguments to be used by the function */
   int *num_arg, /* the number of arguments (in Vensim) the function takes 
                    note that for user loop functions this will be one less
                    than the number of arguments the function actually takes on */
   int *num_vector, /* the number of arguments that are passed as real number vectors */
   int *func_index, /* a number between 0 and 254 that identifies the function
                       vensim_external is called with this number */
   int *dim_act,  /* reserved - for doing dimensional analysis but not implemented */
   int *modify,  /* a flag to indicate that the function will modify value that
                    are passed to it - 
                    0 is a normal function that does not modify its argument
                    1 is a function that does modify arguments
                    2 is a function that modifies arguments and serves as a solver
                      of a simultaneous set of conditions as FIND_ZERO */
  int *num_loop, /* the number of loops that are managed by the function - 
                    this is nonzero (normally 1 or 2) for a function that 
                    needs to return a vector or matrix of values - if this is
                    nonzero Vensim will put a pointer to the vector or array to
                    be filled in and pass it as the first argument to the function 
                    NOTE use -1 for a constdef function and -2 for a datadef function */

  int *num_literal, /* the number of literals that are passed to the function - 
                       arguments are always passed in the order literals, 
                       lookups, vectors, numbers - if num_loop is set the
                       first argument is a vector even if num_literal is positive */
  int *num_lookup  /* the number of lookup functions passed - this structure is
                      not currently accessible but will be made so in the future */
  )
 {

  if(Flist[i].sym) {
     *sym = (unsigned char *)Flist[i].sym ;
     *argument_desc = (unsigned char *)Flist[i].argument_desc ;
     *num_arg = Flist[i].num_args ;
     *num_vector = Flist[i].num_vector ;
     *func_index = Flist[i].func_index ;
     *dim_act = 0 ;
     *modify = Flist[i].modify ;
     *num_loop = Flist[i].num_loop ;
     *num_literal = Flist[i].num_literal ;
     *num_lookup = Flist[i].num_lookup ;
     return(1) ;
     }
  return(0) ; /* indicating the end of the list */
  }

/***************************
 5a some memory management utility routines used by the examples that
    may be of value for other functions - note different routines
    to use Windows GlobalAlloc etc or just calloc etc
  ***************************************/

/* any flags that individual function need set to perform properly on the next
   invocation must be reset by vext_clearmem */
    
static HANDLE *Mem_used = '\0' ;
static int Num_mem_used = 0 ;
static int Max_mem_used = 0 ;


static void *vext_allocate(unsigned nbytes,HANDLE *hndl) 
{
	HANDLE lhndl ;
	if(Num_mem_used >= Max_mem_used) {
		Max_mem_used += 100 ;
		if(Mem_used) {
			Mem_used = (HANDLE *)realloc(Mem_used,Max_mem_used*sizeof(HANDLE)) ;
			memset(Mem_used+Max_mem_used-100,'\0',100*sizeof(HANDLE)) ;
		}
		else
			Mem_used = (HANDLE *)calloc(Max_mem_used,sizeof(HANDLE)) ;
	}
	if(!Mem_used)
		return('\0') ;
	lhndl = (HANDLE)malloc(nbytes) ;
	if(!lhndl)
		return('\0') ;
	if(hndl)
		*hndl = lhndl ;
	Mem_used[Num_mem_used++] = lhndl ;
	return(lhndl) ;
}

static void *vext_reallocate(unsigned nbytes,HANDLE *hndl) 
{
	int i ;
	if(!*hndl)
		return(vext_allocate(nbytes,hndl)) ;
	if(Mem_used) {
		/* find old - otherwise live with the memory leak */
		for(i=0;i<Num_mem_used;i++)
			if(Mem_used[i] == *hndl)
				break ;
	}
	*hndl = realloc(*hndl,nbytes) ;
	if(Mem_used) {
		if(i < Num_mem_used)
			Mem_used[i] = *hndl ;
	}
	return(*hndl) ;
}

static void vext_clearmem()
{
	int i ;

	if(Mem_used) {
		for(i=0;i<Num_mem_used;i++) {
			if(Mem_used[i]) {
				free(Mem_used[i]) ;
			}
		}
		free(Mem_used) ;
	}
	Mem_used = '\0' ;
	Num_mem_used = Max_mem_used = 0 ;

}

/*********************************************************************
 6 - startup and shutdown routines
 *********************************************************************
    these two functions (if they exist and are exported) 
    are called before the simulation starts and
    after it ends - in a normal simulation the order is
       simulation_setup(1) ;
       simulation_setup(0) ;
       simulation_shutdown(0) ;
       simulation_shutdown(1) ;
    for an optimization it the middle two calls are repeated for 
    every simulation - the setup routine should return 0 on failure
    the return value is only used when iniflag == 1.  If the function
    returns 0 simulation will not proceed. 
 ******************************************************************/

CFUNCTION int VEFCC simulation_setup(int iniflag)
 {
  return(1) ;
  }

CFUNCTION int VEFCC simulation_shutdown(int finalflag)
 {
  vext_clearmem() ;
  return(1) ;
  }

/* this is a safety function to validate vector ranges when passing
   vectors to Vensim - you don't need to use it but it will help to 
   prevent nasty memory errors */

static void validate_vector_arg(VECTOR_ARG *v,int firstind,int lastind)
{
   int itemp ;
   if(firstind > lastind) {
      itemp = firstind ;
      firstind = lastind ;
      lastind = itemp ;
      }
   if(v->vals + firstind < v->firstval ||
      v->vals+lastind > v->firstval + v->dim_info->tot_vol) {
      (*VENGV->error_message)(VERROR,(unsigned char *)"Vector argument out of bounds") ;
      v->vals[0] = (COMPREAL)0.0 ;
      v->vals[0] = (COMPREAL)(1.0/v->vals[0]) ; /* generate a floating point exception */
      }
   }


/*********************************************************************
 6 - vensim_external - the actual external function call
 *********************************************************************
   note that all the functions doing floating point are passed and 
   return doubles to prevent any compiler specific problems from 
   arising 
 ******************************************************************/

CFUNCTION int VEFCC vensim_external(VV *val,int nval,int funcid)
 {
  double rval ;

  switch(funcid) {
  case LP_ALLOC_FN:
	  {
		  // *fstack = (COMPREAL) LP_ALLOC( *alloc, *ds_priority, *ds_capacity, *d_capacity, *s_capacity );
		  rval = LP_ALLOC(val[0].vec, // alloc
			  val[1].vec, // ds_priority
			  val[2].vec, // ds_capacity
			  val[3].vec, // d_capacity
			  val[4].vec ); // s_capacity
	  }
	  break;
  case LP_SOLVE_FN:
	  {
		  rval = LP(val[0].vec, // solution
			  val[1].vec, // parameters
			  val[2].vec, 
			  val[3].vec, 
			  val[4].vec,
			  val[5].vec,
			  val[6].vec); 
	  }
	  break;
  default :
	  return(0) ; /* indicate an error condition to Vensim */
  }
  /* set val[0], this value will be used in the equation output */
  val[0].val = (COMPREAL)rval ;
  return(1) ; /* a 1 return value signals vensim of successful completetion */
  }


/*************************************************************************
 7 - actual function bodies - these could be a separate file 
 *************************************************************************
 actual function bodies are all set up to use and return doubles
 (except when acting on vectors) this aids portability across different
 platforms and compilers as the C standard for floating point argument
 passing uses doubles.
 *************************************************************************/

/*************************************
  LP allocation
  ****************************************/

// family of subscript in position n, counting right to left
ST_INDEX ss_family(VECTOR_ARG *vecvar,int n)
{
	if (vecvar->dim_info->tot_dim >= n)
		return vecvar->dim_info->fam[vecvar->dim_info->tot_dim-n];
	else
		return 0;
}

// dimension of subscript in position n, counting right to left
int ss_dim(VECTOR_ARG *vecvar,int n)
{
	if (vecvar->dim_info->tot_dim >= n)
		return vecvar->dim_info->dim[vecvar->dim_info->tot_dim-n];
	else
		return 0;
}


#include "lp.h"

/***************
Allocation with demanders x suppliers, with limited capacity for each demand x supply channel combination
****************/

void do_lp_alloc(COMPREAL *allocation, COMPREAL *ds_priority, COMPREAL *ds_capacity, COMPREAL *d_capacity, COMPREAL *s_capacity, int ndemand, int nsupply);

BigREAL LP_ALLOC( VECTOR_ARG *allocation, VECTOR_ARG *ds_priority, VECTOR_ARG *ds_capacity, VECTOR_ARG *d_capacity, VECTOR_ARG *s_capacity) 
{

   int   nsupply;    /* number of sellers */
   int   ndemand;     /* number of buyers */

   ndemand = ss_dim(allocation,2);
   nsupply = ss_dim(allocation,1);

   if (ndemand < 1 || nsupply < 1)
   {
	   VENGV->error_message( VERROR, "Allocation requires supply and demand dimensions" );
   }

   /* on the first call validate all the dimensions - this is actually done by checking
      families to be sure that seller and buyrer subscripts are used consistently - stronger
      check and if passes no need to test the numbers themselves */
   if (VENGV->time_minus < VENGV->initial_time)
   {
      ST_INDEX cfam = ss_family(allocation,1);
      ST_INDEX dfam = ss_family(allocation,2);

	  if ( cfam != ss_family(ds_priority,1) || cfam != ss_family(ds_capacity,1) || cfam != ss_family(s_capacity,1) )
	  {
		  VENGV->error_message( VERROR, "Supply subscript family mismatch" );
	  }
	  	  if ( dfam != ss_family(ds_priority,2) || dfam != ss_family(ds_capacity,2) || dfam != ss_family(d_capacity,1) )
	  {
		  VENGV->error_message( VERROR, "Demand subscript family mismatch" );
	  }

      if (ds_priority->dim_info->tot_dim < 2)
      {
         VENGV->error_message( VERROR, "Priority must be subscripted [d,s]" );
      }

      if (ds_capacity->dim_info->tot_dim < 2)
      {
         VENGV->error_message( VERROR, "Capacity must be subscripted [d,s]" );
      }

      if (d_capacity->dim_info->tot_dim < 1)
      {
         VENGV->error_message( VERROR, "Capacity must be subscripted [d]" );
      }

      if (s_capacity->dim_info->tot_dim < 1)
      {
         VENGV->error_message( VERROR, "Capacity must be subscripted [s]" );
      }
	  
	  /* todo?
      if (demand_q->u.val + ndemand > demand_q->firstval + demand_q->dim_info->u1.tot_vol)
      {
         market_price_fatal_error( "subscript range error in demand quantity" );
      }

      if (supply_q->u.val + nsupply > supply_q->firstval + supply_q->dim_info->u1.tot_vol)
      {
         market_price_fatal_error( "subscript range error in supply quantity" );
      }
	  */

      /* everything is OK now we actually do something - all errors cause exit */
   }

   /* OK for comps - pass to the raw work function to do actual execution */

   allocation->vals[0] = (COMPREAL) 0.0; /* will be replaced in function call */
   
   do_lp_alloc( allocation->vals, ds_priority->vals,  ds_capacity->vals, d_capacity->vals, s_capacity->vals , ndemand, nsupply);

   return allocation->vals[0];

}

#define MAXCOL 10000

void do_lp_alloc(COMPREAL *allocation, COMPREAL *ds_priority, COMPREAL *ds_capacity, COMPREAL *d_capacity, COMPREAL *s_capacity, int ndemand, int nsupply)
{
	lprec *lp; // lp record
	int i,j,k;
	int c;
	int Ncol; 
	int colno[MAXCOL+1];
	int ret = 0;
	REAL row[MAXCOL+1];
	double obj; // objective function result
	unsigned char msg[512]; // todo: careful not to exceed

	Ncol = ndemand*nsupply; // # of columns (variables)
	if(Ncol>MAXCOL)
	{
		sprintf(msg,"Problem too big - #demand*#supply must be <= %i",MAXCOL); // this is an arbitrary limit and could be revised
		VENGV->error_message( VERROR, msg );
		return;
	}

	lp = _make_lp(0, Ncol);
	if(lp == NULL)
	{
		VENGV->error_message( VERROR, "Failed to create lp_solve model" );
		// todo: set output to NA ?
		return;
	}

	/* Note that it is advised to set the objective function (via set_obj_fn, set_obj_fnex, str_set_obj_fn, set_obj) 
	before adding rows. This especially for larger models. This will be much more performant than adding the objective function afterwards. */

	// objective function
	for (i = 0; i < ndemand; i++)
	{
		for (j = 0; j < nsupply; j++)
		{
			k = i*nsupply + j ;
			row[k+1]=(REAL)ds_priority[k];
		}
	}
	if (! _set_obj_fnex(lp, Ncol, row, NULL ))
	{
		VENGV->error_message( VERROR, "Failed to set objective" );
		return;
	}

	/* set the object direction to maximize */
	_set_maxim(lp);

	_set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

	// sd constraint rows - as bounds
	for (i = 0; i < ndemand; i++)
	{
		for (j = 0; j < nsupply; j++)
		{
			k = i*nsupply + j ;
			if( ds_capacity[k]>=0 )
			{
				if( ! _set_bounds(lp, k+1, 0.0, (REAL)ds_capacity[k]) )
				{
					sprintf(msg,"Failed to set bounds %i %i", i, j);
					VENGV->error_message( VERROR, msg );
					return;
				}

			}
		}
	}

	// d constraint rows
	for (i = 0; i < ndemand; i++)
	{
		if( d_capacity[i]>=0 )
		{
		    c=0;
			for (j = 0; j < nsupply; j++)
			{
				k = i*nsupply + j ;				
				colno[c] = k+1;
				row[c] = 1.0;
				c++;

			}
			if (! _add_constraintex(lp, c, row, colno, LE, (REAL)d_capacity[i]) )
			{
				sprintf(msg, "Failed to add demand constraint %i", i );
				VENGV->error_message( VERROR, msg );
				return;
			}
		} 
	}

	// s constraint rows
	for (j = 0; j < nsupply; j++)
	{
		if( s_capacity[j]>=0 )
		{	
			c=0;
			for (i = 0; i < ndemand; i++)
			{
				k = i*nsupply + j ;				
				colno[c] = k+1;
				row[c] = 1.0;
				c++;
			}
			if ( ! _add_constraintex(lp, c, row, colno, LE, (REAL)s_capacity[j]) )
			{
				sprintf(msg, "Failed to add supply constraint %i", j );
				VENGV->error_message( VERROR, msg );
				return;
			}
		} 
	}

	_set_add_rowmode(lp, FALSE); /* rowmode should be turned off again when done building the model */

	_set_verbose(lp, IMPORTANT);

	ret = _solve(lp);

	switch (ret) {
	case OPTIMAL :
	case PRESOLVED :
		break; // success!
	case SUBOPTIMAL :
		VENGV->error_message( WARNING, "Suboptimal solution" );
		break; // can't happen w/o integer constraints
	case UNKNOWNERROR :
	case DATAIGNORED :
	case NOBFP :
	case NOMEMORY :
	case NOTRUN :
	case INFEASIBLE :
	case UNBOUNDED :
	case DEGENERATE :
	case NUMFAILURE :
	case USERABORT :
	case TIMEOUT :
	case RUNNING :
	default :
		VENGV->error_message( VERROR, "Failed to solve - writing to lp_solve_dump.txt" );
		//#ifdef _DEBUG
	    _write_lp(lp, "lp_solve_dump.txt" );
        //#endif
		return;
	}

#ifdef _DEBUG
	/* objective value */
	obj = _get_objective(lp);
	sprintf(msg,"Objective value: %g\n",obj);
    VENGV->error_message( NOT_ERROR, msg );
#endif

    /* variable values - copy solution */
    _get_variables(lp, row);
    for(k = 0; k < Ncol; k++) {
		//j = k % nsupply;
		//i = k / nsupply;
		//ASSERT(i<=ndemand);
		allocation[k]=(COMPREAL)row[k];
	}

	/* just copy the capacity into the allocation as a test 
	for (i = 0, k = 0; i < ndemand; i++)
	{
		for (j = 0; j < nsupply; j++, k++)
		{
			allocation[k] = ds_capacity[k];
		}
	}
	*/

	if(lp != NULL) {
		/* clean up such that all used memory by lpsolve is freed */
		_delete_lp(lp);
	}

}

/***************
Generic LP
****************/

void do_lp(COMPREAL *solution, COMPREAL *objective, COMPREAL *lbounds, COMPREAL *ubounds, COMPREAL *constraints, COMPREAL *c_vals, COMPREAL *c_types, int nconstr, int nvars);

BigREAL LP( VECTOR_ARG *solution, VECTOR_ARG *objective, VECTOR_ARG *lbounds, VECTOR_ARG *ubounds, VECTOR_ARG *constraints, VECTOR_ARG *c_vals, VECTOR_ARG *c_types ) 
{

   int   nconstr;    /* number of constraints */
   int   nvars;     /* number of variables */

   nvars = ss_dim(constraints,1);
   nconstr = ss_dim(constraints,2);

   if (nvars < 1 || nconstr < 1)
   {
	   VENGV->error_message( VERROR, "Allocation requires constraint and variable dimensions" );
   }

   /* on the first call validate all the dimensions - this is actually done by checking
      families to be sure that seller and buyrer subscripts are used consistently - stronger
      check and if passes no need to test the numbers themselves */
   if (VENGV->time_minus < VENGV->initial_time)
   {
      ST_INDEX cfam = ss_family(constraints,2);
      ST_INDEX vfam = ss_family(constraints,1);

	  if ( cfam != ss_family(c_vals,1) || cfam != ss_family(c_types,1) )
	  {
		  VENGV->error_message( VERROR, "Constraint subscript family mismatch" );
	  }
	  	  
	  if ( vfam != ss_family(solution,1) || vfam != ss_family(objective,1) || vfam != ss_family(lbounds,1) || vfam != ss_family(ubounds,1) )
	  {
		  VENGV->error_message( VERROR, "Variable subscript family mismatch" );
	  }

      if (constraints->dim_info->tot_dim < 2)
      {
         VENGV->error_message( VERROR, "Constraints must be subscripted [c,v]" );
      }

      if (solution->dim_info->tot_dim < 1)
      {
         VENGV->error_message( VERROR, "LHS must be subscripted [v]" );
      }

      if (objective->dim_info->tot_dim < 1)
      {
         VENGV->error_message( VERROR, "Objective must be subscripted [v]" );
      }

      if (lbounds->dim_info->tot_dim < 1)
      {
         VENGV->error_message( VERROR, "Lower bounds must be subscripted [v]" );
      }
	 
      if (ubounds->dim_info->tot_dim < 1)
      {
         VENGV->error_message( VERROR, "Upper bounds must be subscripted [v]" );
      }

      if (c_vals->dim_info->tot_dim < 1)
      {
         VENGV->error_message( VERROR, "Constraint bounds must be subscripted [c]" );
      }

      if (c_types->dim_info->tot_dim < 1)
      {
         VENGV->error_message( VERROR, "Constraint types must be subscripted [c]" );
      }

      /* everything is OK now we actually do something - all errors cause exit */
   }

   /* OK for comps - pass to the raw work function to do actual execution */

   solution->vals[0] = (COMPREAL) 0.0; /* will be replaced in function call */
   
   do_lp( solution->vals, objective->vals,  lbounds->vals, ubounds->vals, constraints->vals, c_vals->vals, c_types->vals, nconstr, nvars);

   return solution->vals[0];

}

#define MAXCOL 10000

void do_lp(COMPREAL *solution, COMPREAL *objective, COMPREAL *lbounds, COMPREAL *ubounds, COMPREAL *constraints, COMPREAL *c_vals, COMPREAL *c_types, int nconstr, int nvars)
{
	lprec *lp; // lp record
	int c,v; //,cv;
	int Ncol; 
	// int colno[MAXCOL+1];
	int ret = 0;
	REAL row[MAXCOL+1];
	double obj; // objective function result
	unsigned char msg[512]; // todo: careful not to exceed

	Ncol = nvars; // # of columns (variables)
	if(Ncol>MAXCOL)
	{
		sprintf(msg,"Problem too big - #variables must be <= %i",MAXCOL); // this is an arbitrary limit and could be revised
		VENGV->error_message( VERROR, msg );
		return;
	}

	lp = _make_lp(0, nvars); // initially 0 constraints, column for each var
	if(lp == NULL)
	{
		VENGV->error_message( VERROR, "Failed to create lp_solve model" );
		// todo: set output to NA ?
		return;
	}

	/* Note that it is advised to set the objective function (via set_obj_fn, set_obj_fnex, str_set_obj_fn, set_obj) 
	before adding rows. This especially for larger models. This will be much more performant than adding the objective function afterwards. */

	// objective function
	for (v = 0; v < nvars; v++)
	{
			row[v+1]=(REAL)objective[v]; // arrays seem to be 1-based?
	}
	if (! _set_obj_fnex(lp, Ncol, row, NULL ))
	{
		VENGV->error_message( VERROR, "Failed to set objective" );
		return;
	}

	/* set the object direction to maximize */
	_set_maxim(lp);

	_set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

	// set bounds
	for (v = 0; v < nvars; v++)
	{
		if( ! _set_bounds(lp, v+1, (REAL)lbounds[v], (REAL)ubounds[v]) )
		{
			sprintf(msg,"Failed to set bound on variable %i", v+1 );
			VENGV->error_message( VERROR, msg );
			return;
		}
	}

	// set constraints
	for (c = 0; c < nconstr; c++)
	{
		if( c_types[c] != NA )               // skip if c_type is :NA:
		{
			int ctype;
			if ( (int)c_types[c] <= 1 )      // LE constraint
				ctype = 1;
			else if ( (int)c_types[c] >= 3 ) // EQ constraint
				ctype = 3;
			else                             // GE constraint
				ctype = 2;

			// wasteful copy, unavoidable because COMPREAL[] is float and REAL[] is double
			// todo: avoid if using DP vensim
			for (v=0; v < nvars; v++)
			{
				double temp = (REAL)constraints[c*nvars+v];
				row[v+1] = (REAL)constraints[c*nvars+v];
			}

			// Ncol = _get_Ncolumns(lp);

			if (! _add_constraint(lp, row, ctype, (REAL)c_vals[c]) )
			{
				sprintf(msg, "Failed to add constraint %i", c+1 );
				VENGV->error_message( VERROR, msg );
				return;
			}
		} 
	}

	_set_add_rowmode(lp, FALSE); /* rowmode should be turned off again when done building the model */

	_set_verbose(lp, IMPORTANT);

			
#ifdef _DEBUG
	    _write_lp(lp, "lp_solve_dump.txt" );
#endif

	ret = _solve(lp);

	switch (ret) {
	case OPTIMAL :
	case PRESOLVED :
		break; // success!
	case SUBOPTIMAL :
		VENGV->error_message( WARNING, "Suboptimal solution" );
		break; // can't happen w/o integer constraints
	case UNKNOWNERROR :
	case DATAIGNORED :
	case NOBFP :
	case NOMEMORY :
	case NOTRUN :
	case INFEASIBLE :
	case UNBOUNDED :
	case DEGENERATE :
	case NUMFAILURE :
	case USERABORT :
	case TIMEOUT :
	case RUNNING :
	default :
		VENGV->error_message( VERROR, "Failed to solve - writing to lp_solve_dump.txt" );
		//#ifdef _DEBUG
	    _write_lp(lp, "lp_solve_dump.txt" );
        //#endif
		return;
	}

#ifdef _DEBUG
	/* objective value */
	obj = _get_objective(lp);
	sprintf(msg,"Objective value: %g\n",obj);
    VENGV->error_message( NOT_ERROR, msg );
#endif

    /* variable values - copy solution */
    _get_variables(lp, row);
    for(v = 0; v < nvars; v++) {

		solution[v]=(COMPREAL)row[v];
	}

	if(lp != NULL) {
		/* clean up such that all used memory by lpsolve is freed */
		_delete_lp(lp);
	}

}